/*
    MODULE: Graphics package for color graphics modes 320x200 and 640x200.
            Code written for Microsoft C Compiler version 3.0 or greater.
            Suggested use is with the 8087 Math Coprocessor.
                                      ---- ---- -----------

        This is simply a two-dimensional graphics package that allows a
        user of MicroSoft C to display graphics.  The package is as
        follows:

        [ 1 ]   GINIT(mode) initializes the graphics enviroment; mode is a
                one letter literal.  The value of mode may be either ( 'H'
                or 'h' ) for 640x200 graphics or ( 'L' or 'l' ) for 320x200
                graphics.  The viewport is initially set to the full size of
                the screen in either (640 or 320)x200 graphics depending upon
                the value of [mode].  The window is initialized with the origin
                at the center of the screen with a range of [-1.0,1.0] in the
                'X' direction and a range of [-1.0,1.0] in the 'Y' direction.
                
                NOTE: All data to be ploted will be mapped through the window
                onto the viewport.  The data must be given in the World
                Coordinate system (i.e.  time starting from  t=0 to lets say
                t=60 microrseconds along the X axis and the volume of air in
                a bursting balloon along the Y axis.).  The data is just mapped
                onto the screen.

        [ 2 ]   GCLEAR() clears the graphics screen.

        [ 3 ]   GCOLOR(color) changes pixel color on future writes to the
                graphics screen.  In High resolution mode, the value of
                [color] must be either a 1 (one) for ON or a 0 (zero) for
                OFF.  In Low resolution mode, the value of [color] may be
                in the range of [0,3] which allows 4 (four) colors: three
                colors and one background.  When GINIT is called, the
                color is initialized to [color=1].

        [ 4 ]   GPLOT(Old_x, Old_y) plots a point in world coordinates
                and uses the current color setting.  The point is mapped
                through the window onto the viewport.  After every plot,
                the current point becomes (Old_x, Old_y).

        [ 5 ]   GDRAW(New_x, New_y) draws a line from the current point to
                the point (New_x, New_y) and uses the current color setting.
                After the line is drawn, the final point becomes the current
                point.

        [ 6 ]   GCIRCLE(center_x, center_y, radius) draws a circle with its
                center at the given point with the given radius.  The point
                and radius are given in world corrdinates.  NOTE:  This
                function plots a circular figure; the circle is modified
                by the size of the viewport (the aspect ratio affects the
                shape).  The function works best if the ranges of the
                horizontal and vertical axis are the same.  The figure is
                plotted in the current color.

        [ 7 ]   GWINDOW(wxmin, wymin, wxmax, wymax) defines the lower left
                and the upper right points of the world coordinates window.

        [ 8 ]   GVIEW(vxmin, vymin, vxmax, vymax) defines the absolute pixel
                values of the upper left and lower right points of the screen.
                NOTE:  For 640x200 graphics. the screen is set up as follows.

                    0,0                                     639,0
                      ******************************************
                      *                                        *
                      *                                        *
                      *                                        *
                      *                                        *
                      *                                        *
                      ******************************************
                    0,199                                   639,199

                    
                and for 320x200 graphics:

                    0,0                                     319,0
                      ******************************************
                      *                                        *
                      *                                        *
                      *                                        *
                      *                                        *
                      *                                        *
                      ******************************************
                    0,199                                   319,199



        
*/

#include <dos.h>
#include <memory.h>
#include <process.h>
#include <ctype.h>
#include <math.h>

typedef unsigned char   UCHAR;
typedef unsigned int    UINT;

static struct GRAPH_MEM    {
    UCHAR  even[8000];
    UCHAR  GAP[0xc0];
    UCHAR  odd[8000];
} SCREEN, *PSCREEN = &SCREEN;

static UCHAR   *peven = SCREEN.even, *podd = SCREEN.odd;

    static      char    ok = 'n';
    static      UINT    x_old, y_old;
    static      UINT    x_new, y_new;
    static      UINT    vxmin, vxmax;
    static      UINT    vymin, vymax;
    static      double  wxmin, wxmax;
    static      double  wymin, wymax;
    static      UINT    Gx,Gy;
    static      char    Gmode;
    static      UCHAR   Gcolor = 1;
    static      UCHAR   hmask[8]= {0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01};
    static      UCHAR   lmask[4]= {0x3f,0xcf,0xf3,0xfc};
    static      UCHAR   lcolor[16] = {  0x00,0x40,0x80,0xc0,
                                        0x00,0x10,0x20,0x30,
                                        0x00,0x04,0x08,0x0c,
                                        0x00,0x01,0x02,0x03 };
    static union REGS   Gregs;

/*****************************************************************************/

static GRAPH_DUMP()
{
    movedata(FP_SEG( PSCREEN ), FP_OFF( PSCREEN ),
        0xb800, 0x00, ( 8000 + 0xc0 + 8000 ) );
    return;

}

/*****************************************************************************/


GINIT(mode)
char mode;
{
    UCHAR   gmode;
    UINT    i;

    if ( mode == 'L' || mode == 'l' ) {
        vxmax = 319;
        gmode = 0x04;
    }
    else if ( mode == 'H' || mode == 'h' ) {
        vxmax = 639;
        gmode = 0x06;
    }
    else {
        printf("\nGINIT: can't open graphics mode: %c\n", mode);
        exit(1);
    }

    Gmode = toupper(mode);

    Gregs.h.ah = 0x00;
    Gregs.h.al = gmode;
    int86(0x10, &Gregs, &Gregs);

    for ( i=0 ; i < 8000 ; i++ )
        SCREEN.even[i] = SCREEN.odd[i] = 0x00;

    GRAPH_DUMP();

    x_old = x_new = y_old = y_new = vxmin = vymin = 0;
    vymax = 199;
    wxmin = wymin = -1.0;
    wxmax = wymax =  1.0;
    Gcolor = 1;

    ok = 'y';

}

/*****************************************************************************/


GCLEAR()
{
    UINT i;

    if ( ok != 'y') {
        printf("\nGCLEAR: graphics not initialized. See GINIT.\n");
        exit(1);

    }

    for ( i = 0 ; i < 8000 ; i++ )
        SCREEN.even[i] = SCREEN.odd[i] = 0x00;
    GRAPH_DUMP();

}

/*****************************************************************************/


GCOLOR(c)
UCHAR c;
{
    if ( ok != 'y') {
        printf("\nGCOLOR: graphics not initialized. See GINIT.\n");
        exit(1);

    }

    if ( Gmode == 'L' ) {
        if ( c < 0 || c > 3 ) {
            printf("\nGCOLOR: color %d not defined",c);
            exit(1);
        
        }
        Gcolor = c;
    }
    else {
        if ( c < 0 ) {
            printf("\nGCOLOR: color %d not defined",c);
            exit(1);
        
        }
        else if ( c == 0 )
            Gcolor = c;
        else
            Gcolor = 1;

    }

}

/*****************************************************************************/


GWINDOW(xmin,ymin,xmax,ymax)
double xmin,ymin,xmax,ymax;
{
    if ( ok != 'y') {
        printf("\nGWINDOW: graphics not initialized. See GINIT.\n");
        exit(1);

    }
    if ( xmin < xmax && ymin < ymax ) {
        wxmin = xmin;
        wymin = ymin;
        wxmax = xmax;
        wymax = ymax;
    }
    else {
        printf("\nGWINDOW: can't open window.\n");
        exit(1);

    }

}

/*****************************************************************************/


GVIEW(xmin,ymin,xmax,ymax)
UINT xmin,ymin,xmax,ymax;
{
    if ( ok != 'y') {
        printf("\nGVIEW: graphics not initialized. See GINIT.\n");
        exit(1);

    }
    if ( xmin < xmax && ymin < ymax && xmin >= 0 && ymin >= 0 &&
         ymax < 200 ) {
        if ( ( Gmode == 'L' && xmax < 320 ) || ( Gmode == 'H' &&
               xmax < 640 ) ) {
            vxmin = xmin;
            vymin = ymin;
            vxmax = xmax;
            vymax = ymax;
        }
    }
    else {
        printf("\nGVIEW: viewport undefined.\n");
        exit(1);
    }

}

/*****************************************************************************/


GPLOT(x,y)
double x,y;
{
    if ( ok != 'y') {
        printf("\nGPLOT: graphics not initialized. See GINIT.\n");
        exit(1);

    }
    if ( (x >= wxmin) && (x <= wxmax) && (y >= wymin) && (y <= wymax) ) {
        Gx = x_old = vxmin + (UINT)( ( ((double)vxmax - (double)vxmin ) /
                    (wxmax - wxmin) ) * ( x - wxmin) ) ;

        Gy = y_old = vymax - (UINT)( ( ((double)vymax - (double)vymin ) /
                    (wymax - wymin) ) * ( y - wymin) ) ;
        Gput();
        GRAPH_DUMP();
    }
    else {
        printf("\nGPLOT: out of range.\n");
        exit(1);
    }

}

/*****************************************************************************/


GDRAW(x,y)
double x,y;
{
    double i,add;
    int   dx,dy,abdx,abdy;

    if ( (x >= wxmin) && (x <= wxmax) && (y >= wymin) && (y <= wymax) ) {
        x_new = vxmin + (UINT)( ( ((double)vxmax - (double)vxmin ) /
                    (wxmax - wxmin) ) * ( x - wxmin) ) ;

        y_new = vymax - (UINT)( ( ((double)vymax - (double)vymin ) /
                    (wymax - wymin) ) * ( y - wymin) ) ;
    }
    else {
        printf("\nGDRAW: out of range.\n");
        exit(1);
    }

    dx = (int)x_new - (int)x_old;
    dy = (int)y_new - (int)y_old;
    abdx = abs(dx);
    abdy = abs(dy);

    if ( abdx == 0 && abdy == 0 )
        return;

    if ( abdx > abdy ) {
        add = 1.0 / (2.0*(double)abdx);
    }
    else {
        add = 1.0 / (2.0*(double)abdy);
    }

    for ( i = 0.0 ; i <= 1.0 ; i += add )  {
        Gx = (UINT) ( (int)x_old + (int)( ((double)dx) * i ) );
        Gy = (UINT) ( (int)y_old + (int)( ((double)dy) * i ) );
        Gput();
    }
    GRAPH_DUMP();

    x_old = x_new;
    y_old = y_new;

}

/*****************************************************************************/


GCIRCLE(x,y,r)
double x, y, r;
{
    double  theta, omega;

    if ( (x+r) > wxmax || (x-r) < wxmin || (y+r) > wymax || (y-r) < wymin ||
             r <= 0.0                                                       ) {
        printf("\nGCIRCLE: out of range.\n");
        exit(1);
    }

    GPLOT(x+r,y);
    for ( theta = 0.0 ; theta <= 360.0 ; theta += 0.5 ) {
        omega = theta * ( 3.14159 / 180.0 );
        GDRAW( x + r*cos(omega), y + r*sin(omega) );
    }

}

/*****************************************************************************/


static Gput()
{
    UCHAR far *pscr;

    if ( Gy % 2 == 0 ) {        /* even vertical address */
        if ( Gmode == 'H') {
            pscr = peven + (Gx / 8) + ( 80 * (Gy >> 1) );
            if ( Gcolor ) {
                *pscr |= hmask[ Gx % 8 ];
            }
            else {
                *pscr &= ~hmask[ Gx % 8 ];
            }
        }
        else {
            pscr = peven + (Gx / 4) + ( 80 * (Gy >> 1) );
            *pscr = ( *pscr & lmask[ Gx % 4 ] ) | 
                            lcolor [ ( ( Gx % 4 ) << 2 ) + Gcolor ];
        }
    }
    else {                      /* odd vertical address */
        if ( Gmode == 'H') {
            pscr = podd + (Gx / 8) + ( 80 * (Gy >> 1) );
            if ( Gcolor ) {
                *pscr |= hmask[ Gx % 8 ];
            }
            else {
                *pscr &= ~hmask[ Gx % 8 ];
            }
        }
        else {
            pscr = podd + (Gx / 4) + ( 80 * (Gy >> 1) );
            *pscr = ( *pscr & lmask[ Gx % 4 ] ) | 
                            lcolor [ ( ( Gx % 4 ) << 2 ) + Gcolor ];
        }
    }
}

/*****************************************************************************/
